package ambientMorality;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.Random;
import java.util.Vector;

import rayTrace.*;
import support.*;

/** Uses the ray tracing code to produce various vignetting curves for KE3. The optical path is described in loadEverything()
 *  and attemptRay() attempts to follow a ray along that path only (it's not a general ray tracing).
 *  
 */

public class AmbientMorality {
	
	final double reHitTolerence = 1e-4;
	final String rootDir = "/work/lidr/rays";
	final String outputDir = rootDir + "/amTest";
	final String opticsFile = rootDir + "/sorted/optics3.txt";
	
	
	Element path[];
	int deathAt[];
	int multiDeath[][];
	
	int svgDrawIfHit = -1; //draw ray into SVG only if it hits something with this surface ID
	
	Collection vertMirrors,horizMirrors,wins1,wins2;
	
	SVGWriter svg = null;
	
	final double origLOS[][] = new double[][]{ {7.8295, 0.0, 0.0}, {1.7342, 0.010, 0.2966 } }; //the original line of sight definition, from centre of input window to beam dump central position
	
	double minTheta=Double.POSITIVE_INFINITY;
	double maxTheta=Double.NEGATIVE_INFINITY;
	double minPhi=Double.POSITIVE_INFINITY;
	double maxPhi=Double.NEGATIVE_INFINITY;
	
	public AmbientMorality(String args[]) {
		
		boolean hotCoords = true;
		double losShift[] = null;
		int nLosPoints = 30;
		double angDefs[][] = new double[][]{ 	{ -4.0, 4.0, 200 }, //min, max and number of phi (angle in y)
												{ -8.0, 2.0, 200 } }; //min,  max and number of theta (angle in Z)
//		double angDefs[][] = new double[][]{ 	{ -10.0, 10.0, 400 }, //min, max and number of phi (angle in y)
//												{ -10.0, 10.0, 400 } }; //min,  max and number of theta (angle in Z)

		for(int i=0;i<2;i++) for(int j=0;j<2;j++) angDefs[i][j] *= Math.PI / 180.0; //convert angles to radians
		
		loadEverything();
				
		double LOS[][] = generateLOSPoints(nLosPoints, 1.5, 4.5, losShift, hotCoords);
		
		//rayDrawings();
		singleOrigScan(LOS,angDefs,"standard.txt");
		//scanVertMirrorTiltZ(LOS,angDefs);
		//scanVertMirrorConverg(LOS,angDefs);
		//scanCh1DetectorPlane(LOS,angDefs);
		//bifocusTheVertMirrors(LOS,angDefs);
		//scanWindowLength(LOS,angDefs);
		
		
		
		if(svg!=null)svg.destroy();
		
		dispMaxAngles();
		
	}
		
	
	/** Similar to scanLOS but we make SVGs */
	public double[][] rayDrawings(){
		double angDefs[][] = new double[][]{ 	{ -4.0, 4.0, 50 }, //min, max and number of phi (angle in y)
												{ -8.0, 2.0, 50 } }; //min,  max and number of theta (angle in Z)
		for(int i=0;i<2;i++) for(int j=0;j<2;j++) angDefs[i][j] *= Math.PI / 180.0; //convert angles to radians
		
		int nPhi = (int)angDefs[0][2];
		int nTheta = (int)angDefs[1][2];
		double dPhi = (angDefs[0][1] - angDefs[0][0])/(nPhi - 1.0);
		double dTheta = (angDefs[1][1] - angDefs[1][0])/(nTheta - 1.0);	

		int nLosPoints = 10;
		double losPoints[][] = generateLOSPoints(nLosPoints, 1.5, 4.5, null, true);
		multiDeath = new int[losPoints.length][];
				
		double losData[][] = new double[nLosPoints][4]; 
		
		svgDrawIfHit = 18;
		
		for(int losIdx=0; losIdx < nLosPoints; losIdx++){
			deathAt = new int[path.length];
			svg = new SVGWriter(outputDir + "/losPoint"+losIdx+".svg");
			
			System.arraycopy(losPoints[losIdx],0,losData[losIdx],0,3); //copy los point into output data
			
			for(int thetaIdx=0;thetaIdx<nTheta;thetaIdx++){
				double theta = angDefs[1][0] + thetaIdx * dTheta;
				for(int phiIdx=0;phiIdx<nPhi;phiIdx++){
					double phi = angDefs[0][0] + phiIdx * dPhi;
					
					Ray ray = createRay(losPoints[losIdx],theta,phi);
					
					if(attemptRay(ray,null,null)){						
						losData[losIdx][3]++;
						checkMaxAngles(theta,phi);
					}
				}
			}

			System.out.println(losIdx + "\t" + losData[losIdx][3]);
			
			multiDeath[losIdx] = deathAt;		
			
			svg.destroy();
		}
		
		
		
		return losData;
	}
	
	
	/** Alex's little idea #564 */
	public void bifocusTheVertMirrors(double LOS[][], double angDefs[][]){
		//double X[] = new double[]{  7.615, 7.715, 7.715, 7.715, 7.615, 7.615 };
		double X[] = new double[]{  7.715, 7.615, 7.615, 7.615, 7.715, 7.715 };
		for(int i=0;i<6;i++){ 
			double A[] = ((Dish)vertMirrors.elems[i]).getCentre();
			
			double N[] = new double[]{ X[i] - A[0], -A[1], -A[2] };
			Ray.reNorm(N);
			
			 ((Dish)vertMirrors.elems[i]).setDishNormal(N);		
		}
		
		singleOrigScan(LOS,angDefs,"bifocal.txt");
		
	}
	
	public void scanWindowLength(double LOS[][], double angDefs[][]){
		int nSteps = 11;
		double start = -0.500;
		double end = 0; //*/
		double dX = (end - start) / ((double)(nSteps-1));
		
		double c2[][] = new double[6][];
		double U[][] = new double[6][3];
		
		//work out the vector from the centre of the first window to the second
		for(int i=0;i<6;i++){
			double c1[] = wins1.elems[i].getCentre();
			c2[i] = wins2.elems[i].getCentre();
			
			double sum = 0;
			for(int j=0;j<3;j++)
				U[i][j] = c2[i][j] - c1[j];
			
			Ray.reNorm(U[i]);
							
		}
		
		for(int posIdx=0;posIdx<nSteps;posIdx++){		
			double x = start + posIdx * dX;
			
			for(int i=0;i<6;i++){
				double c[] = new double[3];
				for(int j=0;j<3;j++)
					c[j] = c2[i][j] + x * U[i][j]; 
					
				wins2.elems[i].setCentre(c);
					
			}
						
			double losData[][] = scanLOS(LOS, angDefs);			
			AsciiMatrixFile.mustWrite(outputDir + "/tubeLen/scan-"+posIdx+".txt",losData);
			
			dispMaxAngles();
		}
	}
	
	/** Produces just a single vignetting curve for the setup as read */
	public void singleOrigScan(double LOS[][], double angDefs[][], String fileName){
		//double losData[][] = scanLOS(LOS, angDefs);
		double losData[][] = scanLOSRandom(LOS, wins1.elems[0],10000);
		
		AsciiMatrixFile.mustWrite(outputDir + "/" + fileName,losData);		
		AsciiMatrixFile.mustWrite(outputDir + "/losDeath.txt",Bits.intToDouble(multiDeath));
	}
	
	/** Produces vignetting curves for a number of different positions of the channel 1 absorption plane (the detector) */ 
	public void scanCh1DetectorPlane(double LOS[][], double angDefs[][]){
		/* int nSteps = 21;
		double start = -3.90;
		double end = -4.20; //*/
		
		int nSteps = 11;
		double start = -3.98;
		double end = -4.02; //*/
		double dX = (end - start) / ((double)(nSteps-1));
		
		Element detectorPlate = Element.findElement(path, 78);
		double x[] = detectorPlate.getCentre();
		
		for(int posIdx=0;posIdx<nSteps;posIdx++){
			x[1] = start + posIdx * dX;
			
			detectorPlate.setCentre(x);
			
			double losData[][] = scanLOS(LOS, angDefs);			
			AsciiMatrixFile.mustWrite(outputDir + "/ch1Plate-fine/scan-"+posIdx+".txt",losData);
			
			dispMaxAngles();
		}
	}
	
	/** Produces the vignetting curve for a number of different tilts of the vertical mirrors
	 *  Each scan points the normal of the dishes at a different point along the y=z=0 line.
	 */
	public void scanVertMirrorTiltZ(double LOS[][], double angDefs[][]){
		int nSteps = 21;
		double span = 0.5 * Math.PI / 180.0;
		double dTheta = span / ((double)(nSteps-1));
		
		
		double c[] = new double[3];
		
		for(int i=0;i<6;i++){
			double x[] = ((Dish)vertMirrors.elems[i]).getCentre();
			for(int j=0;j<3;j++)
				c[j] += x[j];
		}		
		for(int j=0;j<3;j++)
			c[j] /= 6;
		
		for(int i=0;i<6;i++){
			((Dish)vertMirrors.elems[i]).rotateOnZ(c, -span/2);
		}
			
		double bigLosData[][] = new double[LOS.length][10];
		
		for(int i=0;i<LOS.length;i++){
			bigLosData[i][0] = LOS[i][0];
			bigLosData[i][1] = LOS[i][1];
			bigLosData[i][2] = LOS[i][2];
		}
		
						
		for(int posIdx=0;posIdx<nSteps;posIdx++){
			for(int i=0;i<6;i++){
				((Dish)vertMirrors.elems[i]).rotateOnZ(c, dTheta);
					
				path[3] = vertMirrors.elems[i];
				double losData[][] = scanLOS(LOS, angDefs);
				
				for(int l=0;l<LOS.length;l++){
					bigLosData[l][4+i] = losData[l][3];
					bigLosData[l][3] += losData[l][3];
				}	
			}
			
			AsciiMatrixFile.mustWrite(outputDir + "/vertMirrorTiltZ/scan-"+posIdx+".txt",bigLosData);
			//AsciiMatrixFile.mustWrite(outputDir + "/singleLos.txt",losData);
			
			dispMaxAngles();
			
		} //*/
	}
	
	
	/** Produces the vignetting curve for a number of different tilts of the vertical mirrors
	 *  Each scan points the normal of the dishes at a different point along the y=z=0 line.
	 */
	public void scanVertMirrorConverg(double LOS[][], double angDefs[][]){
		int nSteps = 21;
		double span = 0.10;
		double centre = 7.615;
		double dX = span / ((double)(nSteps-1));
		
		for(int i=0;i<6;i++){
			double x[] = ((Dish)vertMirrors.elems[i]).getCurvCentre();
			System.out.println(x[0] + "\t" +x[1] + "\t" + x[2]);
		}
		double convergPoints[][] = generateLOSPoints(nSteps, centre-span, centre+span, null, true);
				
		//for(int posIdx=0;posIdx<nSteps;posIdx++){
		int posIdx = 14; {
			double X = centre - (span / 2.0) + posIdx * dX; 
			
			for(int i=0;i<6;i++){ 
				double A[] = ((Dish)vertMirrors.elems[i]).getCentre();
				
				//double N[] = new double[]{ X - A[0], -A[1], -A[2] };
				double N[] = new double[]{ convergPoints[posIdx][0] - A[0], convergPoints[posIdx][1] - A[1], convergPoints[posIdx][2] - A[2] };
				Ray.reNorm(N);
				System.out.println(N[0] + "\t" +N[1] + "\t" + N[2]);
				
				 ((Dish)vertMirrors.elems[i]).setDishNormal(N);		
				
			}
					
			double losData[][] = scanLOS(LOS, angDefs);			
			//AsciiMatrixFile.mustWrite(outputDir + "/vertMirrorConverg/scan-"+posIdx+".txt",losData);
			AsciiMatrixFile.mustWrite(outputDir + "/singleLos.txt",losData);
			
			dispMaxAngles();
			
		} //*/
	}
	
	
	/** Load the window data and optics from the Optic text file and pieces together the required line of optical components */ 
	public void loadEverything(){
	
		//load the window data, which is seperate
		wins1 = Collection.loadDiscs(rootDir + "/windowDiscsInner.txt", reHitTolerence);	//inner window tube ends
		wins2 = Collection.loadDiscs(rootDir + "/windowDiscsOuter.txt", reHitTolerence);	//out window tube ends
		wins1.ID = 100;
		wins2.ID = 101;
		
		//load the optics data from the text file
		Element elems[] = OpticElementsFile.read("/work/lidr/rays/sorted/optics.txt");
	
		//get the horizontal and verical mirror assemblies as collections so the ray can hit either
		vertMirrors = new Collection( findElements(elems, new int[]{ 2,3,4,5,6,7 }) );
		horizMirrors = new Collection( findElements(elems, new int[]{ 10, 11, 12, 13, 14, 15 }) );
		Element laserBlock = Element.findElement(elems, 8);
		Element laserMirror = Element.findElement(elems, 9);
		vertMirrors.ID = 102;
		horizMirrors.ID = 103;
		
		//Element theRest[] = findElements(elems, new int[]{ 16, 17 }); //only as far as the roof penetrations
		
		//Element theRest[] = findElements(elems, new int[]{ 16, 17, 18, 19 }); //only as far as the lens after the roof penetrations
		
		Element theRest[] = findElements(elems, new int[]{ 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 29, 28, 30, 31, 
															32, //h-alpha/ruby filter 
															31, 30, 69, 68, 65, 68, 69, 66, 67, 70, 71, 72, 73, 74, 75, 76, 78 } ); //and ch1 stuff
		  
		path = new Element[theRest.length + 6];		
		path[0] = wins1;
		path[1] = wins2;
		path[2] = laserBlock;
		path[3] = vertMirrors;		
		path[4] = laserMirror;
		path[5] = horizMirrors;
		System.arraycopy(theRest,0,path,6,theRest.length); //and everything else in order //*/
		
		/*path = new Element[theRest.length + 5];		
		path[0] = wins1;
		path[1] = laserBlock;
		path[2] = vertMirrors.elems[1];		
		path[3] = laserMirror;
		path[4] = horizMirrors;
		System.arraycopy(theRest,0,path,5,theRest.length); //and everything else in order //*/


	}
	
	/** build a collection from the elements in elems with IDs */
	public Element[] findElements(Element elems[], int IDs[]){
		Element collElems[] = new Element[IDs.length];
		
		for(int i=0;i<IDs.length;i++){
			collElems[i] = Element.findElement(elems, IDs[i]);
			if(collElems[i] == null)
				throw new RuntimeException("Failed to find element " + IDs[i] + " in optics file");
		}
		
		return collElems;
	}
	
	
	
	
	/** Attempt to trace ray from the optics. Returns all intersection points with any surface IF target is reached, null if not */
	public boolean attemptRay(Ray ray, String allRaysFile, String completeRaysFile){
		
		Vector<double[]> line = new Vector<double[]>();
		boolean drawThisRay = false;
		
		double x[] = new double[3];
		System.arraycopy(ray.start, 0,x,0,3);
		line.add(x);
		
		for(int i=0;i<path.length;i++){
			
			double hitData[][] = ray.traceRay(path[i], 1, null);
			
			if(hitData == null){  //if it didn't hit the next object
				if(!path[i].isAbsorbing()){ //then abort this ray (unless it was absorbing, then we assume it wasn't supposed to hit it)
					deathAt[i]++;
					if(i>1 && drawThisRay)svg.addLine(line);					
					return false; //with the exception of the last surface, which we should hit!
				}
				else if(i >= (path.length-1)){  //with the exception of the last surface, which should hit!
					if(i>1 && drawThisRay)svg.addLine(line);
					return false;
				}
			}else{		//if we did hit it
				line.add(hitData[1]);
				if(svgDrawIfHit >=0 && path[i].ID == svgDrawIfHit)
					drawThisRay = true;
				
				if(path[i].isAbsorbing()){ //we're ok, unless it was abosrbing, then we assume it wasn't supposed to, so give up on the ray
					deathAt[i]++;
					if(i>0 && drawThisRay)svg.addLine(line);
					return (i >= (path.length-1)); //with the exception of the last surface, which we should hit and thats the ray successful!
				}
			}
			
			
		}
		
		if(drawThisRay)
			svg.addLine(line);
		return true;
		
	}
	
	/** Work along the points provided in losPoints, running angles defined by angDefs through optics and returning the number of completem rays made on each */
	public double[][] scanLOS(double losPoints[][], double angDefs[][]){
		multiDeath = new int[losPoints.length][];
		
		int nPhi = (int)angDefs[0][2];
		int nTheta = (int)angDefs[1][2];
		double dPhi = (angDefs[0][1] - angDefs[0][0])/(nPhi - 1.0);
		double dTheta = (angDefs[1][1] - angDefs[1][0])/(nTheta - 1.0);	

		int nLosPoints = losPoints.length;
		
		
		double losData[][] = new double[nLosPoints][4]; 
		
		for(int losIdx=0; losIdx < nLosPoints; losIdx++){
			System.arraycopy(losPoints[losIdx],0,losData[losIdx],0,3); //copy los point into output data
			deathAt = new int[path.length];
			
			for(int thetaIdx=0;thetaIdx<nTheta;thetaIdx++){
				double theta = angDefs[1][0] + thetaIdx * dTheta;
				for(int phiIdx=0;phiIdx<nPhi;phiIdx++){
					double phi = angDefs[0][0] + phiIdx * dPhi;
					
					Ray ray = createRay(losPoints[losIdx],theta,phi);
					
					
					if(attemptRay(ray,null,null)){						
						losData[losIdx][3]++;
						checkMaxAngles(theta,phi);
						
					}
				}
				//System.out.println(losIdx + "\t" + theta + "\t" + losData[losIdx][3]);
				
			}

			//return losData;
			System.out.println(losIdx + "\t" + losData[losIdx][3]);
			//dispDeathSurfaces();
			multiDeath[losIdx] = deathAt;			
			
		}
		return losData;
	}
	
	public double[][] scanLOSRandom(double losPoints[][], Element target, int nRays){
		multiDeath = new int[losPoints.length][];
		
		int nLosPoints = losPoints.length;
		
		
		double losData[][] = new double[nLosPoints][4]; 
		
		for(int losIdx=0; losIdx < nLosPoints; losIdx++){
			System.arraycopy(losPoints[losIdx],0,losData[losIdx],0,3); //copy los point into output data
			deathAt = new int[path.length];
			
			for(int i=0;i<nRays;i++){
					
				//Ray ray = createRay(losPoints[losIdx],theta,phi);
				Ray ray = Ray.generateRandomRayTowardSurface(losPoints[losIdx], target);
						
				if(attemptRay(ray,null,null))						
					losData[losIdx][3]++;					
				
			}
			
			

			//return losData;
			System.out.println(losIdx + "\t" + losData[losIdx][3]);
			//dispDeathSurfaces();
			multiDeath[losIdx] = deathAt;			
			
		}
		return losData;
	}
	
	/** Create a set number of points along the original line of sight, shifted at the beamdump end by shift (if not null) */
	public double[][] generateLOSPoints(int nPoints, double R0, double R1,double shift[], boolean hot){
		
		double U[] = new double[3];//unit vector along LOS
		double losStart[]  = new double[3]; //start of LOS points
		double losEnd[]  = new double[3]; //end of LOS points
		double losLen[]  = new double[3]; //length of LOS
		double losStep[]  = new double[3];//step along LOS
		
		double thisLOS[][] = new double[2][3];
		
		
		//first we need to sort out the start and end of the LOS
		for(int i=0;i<3;i++){
			thisLOS[0][i] = origLOS[0][i];
			thisLOS[1][i] = origLOS[1][i];
			if(hot){ //apply correction for temperature
				thisLOS[0][i] *= 1.0022;
				thisLOS[1][i] *= 1.0022;
			}
			if(shift != null) //apply any shift to beamDump end
				thisLOS[0][i] += shift[i];
		
			U[i] = thisLOS[1][i] - thisLOS[0][i]; //unit vec is end-start 
			losStart[i] = thisLOS[0][i] + U[i] * ( (R0 - thisLOS[0][0]) / U[0]);
			losEnd[i] = thisLOS[0][i] + U[i] * ( (R1 - thisLOS[0][0]) / U[0]);
			losLen[i] = losEnd[i] - losStart[i];
			losStep[i] = losLen[i] / (nPoints-1);
		}
		
		double LOS[][] = new double[nPoints][3];
		
		for(int i=0;i<nPoints;i++)
			for(int j=0;j<3;j++)
				LOS[i][j] = losStart[j] + i * losStep[j];
		
		return LOS;
		
		
	}
	
	
	/** Create a ray, starting from point losIdx on the current LOS */ 
	public Ray createRay(double start[], double theta, double phi){
		double unit[] = new double[]{ Math.cos(phi)*Math.cos(theta),
										Math.sin(phi),
										Math.cos(phi)*Math.sin(theta) };		
		return new Ray(start,unit);		
	}
	
	public void checkMaxAngles(double theta, double phi){
		if(theta<minTheta)minTheta = theta;
		if(theta>maxTheta)maxTheta = theta;
		if(phi<minPhi)minPhi = phi;
		if(phi>maxPhi)maxPhi = phi;
	}
	
	public void dispMaxAngles(){
		System.out.println("minTheta = " + minTheta*180/Math.PI);
		System.out.println("maxTheta = " + maxTheta*180/Math.PI);
		System.out.println("minPhi = " + minPhi*180/Math.PI);
		System.out.println("maxPhi = " + maxPhi*180/Math.PI);
	}
	
	public void dispDeathSurfaces(){
		for(int i=0;i<path.length;i++){
			System.out.println(i + " ("+path[i].ID+"): " + deathAt[i]);
			deathAt[i]=0;	
		}
	}
	
		
	public static void main(String[] args) { new AmbientMorality(args); }

	
	
}
