package rayTrace;

import java.util.Random;



public class Ray {
	
	public Ray(double start[], double unit[]) {
		System.arraycopy(start,0,this.start,0,3);
		System.arraycopy(unit,0,this.unit,0,3);		
	}
	
	public double start[] = new double[3];
	public double unit[] = new double[3];
	
	public void reflectAt(Intersection hit){
		double c = 0;
		for(int i=0;i<3;i++){
			start[i] = hit.pos[i];
			c += hit.normal[i] * unit[i];
		}
		
		for(int i=0;i<3;i++)
			unit[i] = unit[i] - (2 * hit.normal[i] * c);
		
		//System.out.println("Rflec:" + Math.sqrt(unit[0]*unit[0] + unit[1]*unit[1] + unit[2]*unit[2]));
		reNormUnit();
	}
	
	public void refractAt(Intersection hit){
		
		//the ray now starts from here
		for(int i=0;i<3;i++) 
			start[i] = hit.pos[i];
		
		//we need to use the object we hit's refractive index ratio to possible refract the ray
		double dotProd=0;
		for(int i=0;i<3;i++)
			dotProd += unit[i] * hit.normal[i];
		
		if( (1 - dotProd*dotProd) >= 1/(hit.rIndexRatio*hit.rIndexRatio)){
			//total internal refraction
			reflectAt(hit);
			return;
		}
		
		double a = - hit.rIndexRatio * dotProd;
		double b = - Math.sqrt( 1 - hit.rIndexRatio*hit.rIndexRatio*(1-dotProd*dotProd));
		double c = a + b;
		
		double newUnit[] = new double[3];
		
		for(int i=0;i<3;i++)
			newUnit[i] = c * hit.normal[i] + hit.rIndexRatio * unit[i];
		
		unit = newUnit;

		//System.out.println("Rfrac:" + Math.sqrt(newUnit[0]*newUnit[0] + newUnit[1]*newUnit[1] + newUnit[2]*newUnit[2]));
		reNormUnit();
	}
	
	public void reNormUnit(){
		unit = reNorm(unit);
	}
	
	public static double[] reNorm(double x[]){
		double sum=0;
		
		for(int i=0;i<3;i++)
			sum += x[i] * x[i];
		
		sum = Math.sqrt(sum);
		
		for(int i=0;i<3;i++)
			x[i] /= sum;
		
		return x;
	}
	
	public static double[] cross(double A[], double B[]){
		double AxB[] = new double[3];
		
		AxB[0] =  A[1] * B[2] - A[2] * B[1];
		AxB[1] = -A[0] * B[2] + A[2] * B[0];
		AxB[2] =  A[0] * B[1] - A[1] * B[0];
		
		return AxB;

	}
	
	public void applyIntersection(Intersection hit){
		
		if(hit.object.isReflective()) //if transparent, just move start pos up to intersection point
			reflectAt(hit);
		else if(hit.object.isTransparent())
			refractAt(hit);			
		else if(hit.object.isAbsorbing()) //otherwise is unknown or absorbing, kill the ray
			unit = null;
		else
			throw new RuntimeException("Unknown material type on element with ID " + hit.object.ID);
		
	}	
		
	/** Traces this ray to/through through a given element or collection */
	public double[][] traceRay(Element elem, int nHits, Element target){
		
		boolean hasHitTarget = false;
		
		double hits[][] = new double[nHits+1][];
		
		hits[0] = new double[4]; 
		System.arraycopy(start,0,hits[0],0,3);
		hits[0][3] = -1;
		int j=1;
		
		for(int i=0;i<(nHits);i++){
			if(unit == null) //a null unit vector means this ray has been killed
				break;
				
			Intersection hit = elem.findIntersection(this);
			if(hit != null){
				applyIntersection(hit);
				//System.out.println(hit.object.ID);
				hits[j] = new double[4];
				System.arraycopy(start,0,hits[j],0,3);
				hits[j][3] = hit.object.ID;
				j++;
				
				if(hit.object == target)
					hasHitTarget = true;				
			}
			
		}	
		
		if(j<=1) //if we never hit ANYTHING, fail
			return null;
		
		if(target != null && !hasHitTarget) //if there was a target, and we didn't hit it
			return null;
		
		return hits;
	}
	
	/** WARNING: this doesn't generate properly if the object is close to the starting point.
	 * The distribution will not be constant over solid angle */
	public static Ray generateRandomRayTowardSurface(double start[], Element target){
		double N[] = new double[3];
		
		double R = target.boundingSphereRadius();
		
		// we generate a random point on the disc presented to the ray by the bounding sphere of the target object
		Random randGen = new Random();
		
		//generate a random r and theta and convert to an x and y on the disc (a and b)
		double r = R * Math.sqrt(randGen.nextDouble()); //this re-biases the radius so that the distribution comes out equal on the disc... somehow
		double theta = randGen.nextDouble() * 2 * Math.PI;
		double a = r * Math.sin(theta);
		double b = r * Math.cos(theta);
		
		double C[] = target.getCentre();
		//generate the vector from the source to the centre of the target		
		for(int i=0;i<3;i++)
			N[i] = C[i] - start[i];
			
		Ray.reNorm(N); //make it a unit
		
		//we now need a perpendicular to this
		double V1[];
		
		//first vector in disc is perp to N and anything
		V1 = Ray.cross(N, new double[]{1,0,0} );
		if(V1[0]*V1[0] + V1[1]*V1[1] + V1[2]*V1[2] == 0)
			V1 = Ray.cross(N, new double[]{0,1,0} );
		
		Ray.reNorm(V1);
		
		//second is perp to N and V1
		double V2[] = Ray.cross(V1, N);
		Ray.reNorm(V2);
		
		double U[] = new double[3]; //unit vector from source to point on target
		
		for(int i=0;i<3;i++)
			U[i] = (C[i] + a * V1[i] + b * V2[i]) - start[i];
		Ray.reNorm(U);
		
		return new Ray(start,U);
		
	}
}
